package at.ac.uibk.gitsearch.repository.search.testESService;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import org.apache.http.HttpHost;
import org.apache.http.util.EntityUtils;
import org.apache.logging.log4j.LogManager;
import org.elasticsearch.action.admin.indices.get.GetIndexRequest;
import org.elasticsearch.action.admin.indices.get.GetIndexResponse;
import org.elasticsearch.action.index.IndexRequest;
import org.elasticsearch.client.RequestOptions;
import org.elasticsearch.client.Response;
import org.elasticsearch.client.RestClient;
import org.elasticsearch.client.RestClientBuilder;
import org.elasticsearch.client.RestHighLevelClient;
import org.elasticsearch.common.settings.Settings;
import org.elasticsearch.node.InternalSettingsPreparer;
import org.elasticsearch.node.Node;
import org.elasticsearch.node.NodeValidationException;
import org.elasticsearch.plugins.Plugin;
import org.elasticsearch.transport.Netty4Plugin;
import org.slf4j.LoggerFactory;
import org.springframework.boot.context.event.ApplicationStartingEvent;
import org.springframework.context.event.ContextStartedEvent;
import org.springframework.context.event.ContextStoppedEvent;
import org.springframework.context.event.EventListener;
import org.springframework.data.elasticsearch.client.NodeClientFactoryBean;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpHeaders;
import org.springframework.http.MediaType;
import org.springframework.web.client.RestClientException;
import org.springframework.web.client.RestTemplate;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.JsonMappingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.google.common.base.Charsets;
import com.google.common.io.ByteSource;

public class ElasticSearchTestServerConfiguration {
	
	private static final org.slf4j.Logger LOGGER = LoggerFactory.getLogger(ElasticSearchTestServerConfiguration.class);
    private TestESNode testNode;

    private ElasticSearchTestServerConfiguration()  {
    }
    
    private static ElasticSearchTestServerConfiguration instance = null;
    
    public static  ElasticSearchTestServerConfiguration getTestInstance()  {
    	if (instance==null)
    		instance = new ElasticSearchTestServerConfiguration();
    	return instance;
    	
    }

	public synchronized void startTestNode() throws IOException, NodeValidationException {
		if(testNode==null) {
			LOGGER.info("Starting ES Test Node");
			try {
			File f = File.createTempFile("tobedeleted", "txt");
			Settings s =  Settings.builder().put("http.port", "29200")
						 .put("transport.port", "29300")
			             .put("node.name", "esSharingTest")
	                     .put("transport.type", "netty4")
		                 .put("http.type", "netty4")
			             .put("path.home", f.getParent() + File.separator + "esTest").build();
			
			NodeClientFactoryBean fb = new NodeClientFactoryBean(true);
			
			testNode = new TestESNode(s, Collections.singleton(Netty4Plugin.class)) {
				
			}; 
			testNode.start();
			
			setUpMetaDataIndex();
			setUpContent();
			LOGGER.info("Started ES Test Node");
			} catch (Throwable t) {
				LOGGER.info("ES Test Node crashed: ", t);
			}
		} else {
			LOGGER.info("Node already started");
		}
	}
	
	private void setUpMetaDataIndex() throws IOException {
		
		RestTemplate restTemplate = new RestTemplate();
		HttpHeaders headers = new HttpHeaders();
	    headers.setContentType(MediaType.APPLICATION_JSON);

	    List<String> indices2 = getIndices();

	    indices2.forEach((s) -> {
	    	LOGGER.info("Found index {}", s);
		    try {
				restTemplate.delete("http://localhost:29200/" + s, String.class); // just in case, delete old index
			} catch (RestClientException e) {
				LOGGER.info("Deletion of previous index failed: {}!", e.getMessage());
			}
	    });
	    
	    
	    
	    
//	    GetIndexRequest request2 = new GetIndexRequest();
//	 
//	    GetIndexResponse response = client.indices().get(request2, RequestOptions.DEFAULT);
//	    String[] indices = response.getIndices();
	    
	    
	    
	    HttpEntity<String> request = 
	    	      new HttpEntity<String>(getMetaDataConfigDefinition(), headers);
	    
	    
	    
	    restTemplate.put("http://localhost:29200/idx_metadata_1", request, String.class);
	    
	}

	private List<String> getIndices() throws IOException, JsonProcessingException, JsonMappingException {
		List<String> indices2 = null;
	    RestHighLevelClient client = new RestHighLevelClient(RestClient.builder(new HttpHost("localhost", 29200)));

	    RestClient restClient = client.getLowLevelClient();
	    Response response2 = null;
	    try {
	        response2 = restClient.performRequest("GET", "/_cat/indices?v&format=json");
	    } catch (IOException e) {
	        LOGGER.warn(e.toString(), e);
	    }

	    // parse the JSON response
	    List<HashMap<String, String>> list = null;
	    if (response2 != null) {
	    	ObjectMapper mapper = new ObjectMapper();
	        String rawBody = EntityUtils.toString(response2.getEntity());
	        TypeReference<List<HashMap<String, String>>> typeRef = new TypeReference<List<HashMap<String, String>>>() {};
	        list = mapper.readValue(rawBody, typeRef);
	    }

	    // get the index names
	    if (list != null) {
	        indices2 = list.stream()
	            .map(x -> x.get("index"))
	            .collect(Collectors.toList());
	    }
		return indices2;
	}
	
	private void setUpContent() throws IOException {
		

		RestTemplate restTemplate = new RestTemplate();
		HttpHeaders headers = new HttpHeaders();
	    headers.setContentType(MediaType.APPLICATION_JSON);
	    
	    int contentCount = 0;
	    
	    while(true) {
	    	contentCount++;
		    final String content = getContent(contentCount);
		    if(content==null) break;
			HttpEntity<String> request = 
		    	      new HttpEntity<String>(content, headers);
		    restTemplate.put("http://localhost:29200/metadata/_doc/"+contentCount, request, String.class);
	    }
	}
	private String getMetaDataConfigDefinition() throws IOException {
		InputStream metaDataStream = this.getClass().getResourceAsStream("../testData/es_metadata.schema.json");
		ByteSource byteSource = new ByteSource() {
			
			@Override
			public InputStream openStream() throws IOException {
				return metaDataStream;
			}
		};
		String text = byteSource.asCharSource(Charsets.UTF_8).read();
		return text;
	}
	
	private String getContent(int i) throws IOException {
		InputStream metaDataStream = this.getClass().getResourceAsStream("../testData/content"+i+".json");
		if(metaDataStream==null) return null;
		ByteSource byteSource = new ByteSource() {
			
			@Override
			public InputStream openStream() throws IOException {
				return metaDataStream;
			}
		};
		String text = byteSource.asCharSource(Charsets.UTF_8).read();
		return text;
	}

    
    @EventListener
    public void onApplicationEvent(ApplicationStartingEvent event) {
		try {
			startTestNode();
		} catch (IOException | NodeValidationException e) {
			LOGGER.error("Cannot startup TestNode", e);
		}
    }

	@EventListener
    public void onApplicationEvent(ContextStartedEvent event) {
		try {
			startTestNode();
		} catch (IOException | NodeValidationException e) {
			LOGGER.error("Cannot startup TestNode", e);
		}
    }
    
    @EventListener
    public void onApplicationEvent(ContextStoppedEvent event) {
		try {
			testNode.close();
		} catch (IOException  e) {
			LOGGER.error("Cannot startup TestNode", e);
		}
    }

    public static class TestESNode extends Node {
        public TestESNode(Settings preparedSettings, Collection<Class<? extends Plugin>> classpathPlugins) {
            super(InternalSettingsPreparer.prepareEnvironment(preparedSettings, null), classpathPlugins, true);
        }

		@Override
		protected void registerDerivedNodeNameWithLogger(String nodeName) {
			LogManager.getLogger(nodeName);
		}
    }
    
}
